library(BRugs)

# ----------------------------------------------------------------------
#     sdose      --> sdose levels
#     target.tox --> target toxcity 
#     ff         --> integer value describing the functional form for the
#                   density of probability :
#                   1 - Hyperbolic tangent
#                   2 - logistic
#                   3 - Power
#     prior.theta--> vector containing the information to generate the
#                   prior distribution of parameter theta
#                   [1] - the prior distribution type:
#                       1 - gamma: mean=a*b; var=a*b*b
#                       2 - uniform: a+b*unif(0,1)
#                   [2] - a: first parameter of the prior distribution
#                   [3] - b: second parameter of the prior distribution
#     alpha          --> generate (1-alpha) of the credible level
#     burnin.itr     --> Number of burn-in iterations in MCMC
#     production.itr --> Number of production iterations in MCMC
#     new            --> TRUE if it is the beginning of an experiment, else FALSE
#     trial.history  --> name of the file where the history of the experiment is
#                         saved into
#     trial.output   --> name of the file where the result of the experiment
#                        is written into
#
# ----------------------------------------------------------------------
crm.interactive <- function(sdose, target.tox, ff = 1, prior.theta =c(1,1,1), alpha=0.05, burnin.itr = 2000,
                           production.itr = 10000, new = 1, trial.history="history.txt", trial.output="output.txt")
{
    k <- length(sdose )   
    # Get prior sampling data
    sampled.theta <- getprior(prior.theta, production.itr)    
    if ( new ) {
        # Beginning of a study
        toxicity <- rep( 0, k )
        No.toxicity <- rep( 0, k )
        npac <- 1
    }
    else {
        # Continuation of the study
        data <- scan(trial.history )
        toxicity <- data[1:k]
        No.toxicity <- data[(k+1):(2*k)]
        npac <- sum(toxicity) + sum(No.toxicity) + 1
    }
    par( mfrow=c(1,2) )        
    # main loop
    repeat {
        cat("\n ***** START SEQUENCE ", npac, "*****\n")
        plot( density(sampled.theta), xlab="theta", ylab="density",
                type="l", main="Prior/Posterior Distribution")       
        # ix is the proposed dose level at which the j-th patient is treated        
        ans <- nextdose(npac, sdose, ff, target.tox, sampled.theta)        
        ix <- ans[1]
        mean.theta <- ans[2]        
        repeat {
            cat("\n\n RECOMMENDED DOSE LEVEL FOR PATIENT ", npac, "IS:", ix)
            ans <- get.dose(k)            
            if (ans==-2)
                ans <- ix            
            if (ans==-1)
                cat("\n\n EXIT WITHOUT SAVING THE RESULTS")
            else if (ans==0)
                cat("\n\n EXIT AND SAVE THE RESULTS SO FAR")            
            if (ans<1) {
                cat("\n\n DO YOU REALLY WANT TO EXIT ? (Y/N)  ")
                yn <- readline()
                if (yn=="y" || yn=="Y") break
                else                    next
            }            
            # y.j is the outcome ( 0 or 1 ) from the treatment of
            # a patient at dose level ix
            cat("\n ENTER TOXICITY OUTCOME FOR PATIENT ", npac)
            y.j <- get.answer( )            
            # give the user a last chance to modify the treatment and outcome
            cat("\n\n\t\t ENTERED VALUES:")
            cat("\n DOSE LEVEL ...", ans)
            cat("\n TOXICITY .....")
            if (y.j) cat(" YES") else cat(" NO")
            cat("\n PRESS `RETURN' IF OK, OR ANY OTHER KEY TO ENTER NEW VALUES ")
            key <- readline()
            if (nchar(key)==0) break
        }        
        if (ans==-1) return (invisible())
        else if (ans==0)  break        
        ix <- ans        
        # update the arrays used in computing the likelihoods
        toxicity[ix] <- toxicity[ix] + y.j
        No.toxicity[ix] <- No.toxicity[ix] + 1 - y.j
        # compute the posterior probabilities
        sampled.theta <- Posterior(ff, prior.theta, toxicity, No.toxicity, sdose, burnin.itr, production.itr)
        npac <- npac + 1
    }    
    # save data for study continuation
    save.data(toxicity, No.toxicity, trial.history)    
    # write the results in a file on the disk
    WriteTable( toxicity, No.toxicity, mean.theta, trial.output )    
    # compute the credible intervals
    cred.int(sdose, alpha, sampled.theta, mean.theta,  ff, target.tox, trial.output )
}

# ----------------------------------------------------------------------
#     User inputs the OUTCOME from the treatment at a dose level:
#     0 - no toxicity
#     1 -    toxicity
# ----------------------------------------------------------------------
get.answer <- function( ) {   
    repeat {
        cat("\n ( 0 - no toxicity, 1 - toxicity )   ")
        ans <- as.numeric(readline())
        if (is.na(ans)) next
        if (ans != floor(ans)) next
        if (ans==0 || ans==1) return(ans)
    }
}

# ----------------------------------------------------------------------
#     ask the user to input an integer number <= n.
#     the number is checked to belong to [-1,n] and also to be
#     an integer.  `ENTER' returns to the caller with no action,
#
#     n - biggest number to be accepted
# ----------------------------------------------------------------------
get.dose <- function( n ) {
    repeat {
        cat("\n\n ENTER DOSE LEVEL BETWEEN 1 AND ", n)
        cat("\n (`RETURN' TO ACCEPT RECOMMENDATION, -1 TO EXIT, 0 TO EXIT AND SAVE DATA)  ")
        
        ans <- readline()
        if ( nchar(ans)==0 ) return( -2 )
        ans <- as.integer(ans)
        if (is.na(ans)) next
        if ( -1<=ans && ans<=n ) return( ans )
    }
}


# ----------------------------------------------------------------------
# generates a vector of values and prior distribution
#
#     prior.theta --> vector containing the information for prior
#               [1] - the prior distribution type:
#                     1 - gamma: mean=a*b; var=a*b*b
#                     2 - uniform: a+b*unif(0,1)
#               [2] - a: first parameter of the prior distribution
#               [3] - b: second parameter of the prior distribution
#    
# ----------------------------------------------------------------------
getprior <- function(prior.theta, n) {
    type <- prior.theta[1]
    a    <- prior.theta[2]
    b    <- prior.theta[3]    
    if ( type==1 ) {
        prior<-rgamma(n,a)*b        
    }
    else if ( type==2 ) {
        prior<-a+runif(n)*b
    }
    return (prior)
}


# ----------------------------------------------------------------------
#     returns the vector containing sampled data from winbug
#
#     ff          --> the model applied in the study
#                    1 - hyperbolic tangent
#                    2 - logistic
#                    3 - Power
#     prior.info  --> vector of prior distribution of parameter theta
#     sdose       --> vector containing the dose level
#     tox         --> number of successes (toxicities)
#     notox       --> number of failed patients (no-toxicities)
#     burnin.itr  --> number of burn-in iterations
#     production.itr --> number of production iterations
# ----------------------------------------------------------------------
Posterior <- function(ff, prior.info, tox, notox, sdose, burnin.itr, production.itr)
{    
    all.patient <- tox + notox
    datan <-all.patient[all.patient!=0]
    datas <-tox[all.patient!=0]
    datad <-sdose[all.patient!=0]
    k <- length(datan)
    if (k == 1)
    {
        datan <- c(datan, 0)
        datas <- c(datas, 0)
        datad <- c(datad, 0)
    }
    mydata <- list(N1 = k, s = datas,n = datan,d = datad, low = prior.info[2], high = prior.info[3])
    bugsData(mydata, file="data.txt")    
    if (prior.info[1] == 1)
    {
        if (ff == 1)
            modelCheck("HTGamma.txt")
        else if (ff == 2)
            modelCheck("LogisticGamma.txt")
        else if (ff == 3)
            modelCheck("PowerGamma.txt")
    }
    else if (prior.info[1] == 2)
    {
        if (ff == 1)
            modelCheck("HTUnif.txt")
        else if (ff == 2)
            modelCheck("LogisticUnif.txt")
        else if (ff == 3)
            modelCheck("PowerUnif.txt")
    }        
    modelData("data.txt")
    modelCompile()
    modelInits("initdat.txt")
    #modelGenInits()
    seed1 <- round(runif(1)*10000)
    modelSetSeed(seed1)
    modelUpdate(burnin.itr)
    samplesSet(c("theta"))
    modelUpdate(production.itr)    
    t<- samplesSample("theta")
    return(t)
}


# ----------------------------------------------------------------------
#     find the new dose and posteria mean of theta
#
#     npac     --> the patient currently being treated
#     sdose     --> vector of allowable dose values
#     ff       --> integer value:
#                    1 - hyperbolic tangent
#                    2 - logistic
#                    3 - Power
#     target.tox    --> desired probability of event (toxicity)
#
#     sample.theta  --> sampled data of variable theta
#
# ----------------------------------------------------------------------
nextdose <- function(npac, sdose, ff, target.tox, sample.theta) {
    k <- length( sdose )    
    mean.theta <- mean(sample.theta)
    tick <- round(as.single(sdose),2)
    t1<-rep(0,k)
    for(i in 1:k)
    {
        if ( ff == 1 )         
            t1[i] <- mean((1+exp(-2*sdose[i]))^(-sample.theta))
        else if ( ff == 2 )
            t1[i] <- mean(1/(1+exp(-3-sample.theta*sdose[i])))
        else if ( ff == 3)
            t1[i] <- mean(sdose[i]^exp(sample.theta))
    }        
    plot( sdose, t1, ylim=c(0,1), pch="*", xlab="sdose",
            ylab="probability of toxicity", axes=F, main="Dose-Toxicity Curve" )
    axis( 1, tick, as.character(tick) )
    axis( 2, las=1 )
    box( )
    abline( target.tox, 0, lty=2 )        
    if (npac==1)
        cat("\n MEAN VALUE FOR PARAMETER \"Theta\" UNDER PRIOR DISTRIBUTION = ")
    else
        cat("\n UPDATED MEAN VALUE FOR PARAMETER \"Theta\" WITH ", npac-1, " PATIENTS ENTERED = ")    
    cat( round( mean.theta, 2 ) )    
    # choose the closest dose to x0
    ix <- order( abs(t1-target.tox))[1]    
    return(c(ix, mean.theta))
}

# ----------------------------------------------------------------------
#
# Saves data for study continuation
#
#     toxicity      --> array containing the number of recorded toxicities
#                     at each dose level.
#     No.toxicity   --> array containing the number of recorded
#                     non-toxicities at each dose level.
#     trial.history --> the name of the file where the above arrays will
#                     be saved for study continuation
# ----------------------------------------------------------------------
save.data <- function( toxicity, No.toxicity, trial.history) {   
    k <- length(toxicity)
    write( toxicity, trial.history, ncol=k )
    write( No.toxicity, trial.history, ncol=k, append=T )
}

# -----------------------------------------------------------------------
#     write in a file, on the same line, text,
#     followed by numeric fields in a required format
# -----------------------------------------------------------------------
output <- function( file.name, txt, num, form, coda, sep ) {
    k <- length( num )
    chnum <- as.character( c( num, form ) )
    txt.out <- c( txt, format( chnum ) )
    txt.out <- c( txt.out[0:k+1], coda)
    write( txt.out, file.name, ncol=50, append=T )
    write( sep, file.name, append=T )
}

# ----------------------------------------------------------------------
#     toxicity    --> array containing the number of recorded toxicities
#                      at each dose level.
#     No.toxicity --> array containing the number of recorded
#                     non-toxicities at each dose level.
#     mean.theta      --> latest mean value of theta
#     trial.output    --> the name of the file where the summary of the study
#                     will be recorded
# ----------------------------------------------------------------------
WriteTable <- function( toxicity, No.toxicity, mean.theta, trial.output ) {
    k <- length( toxicity )
    sep <- "|-----------------------------------------------------------|"
    form <- "1234"
    coda <- " |"
    write( sep, trial.output )
    subjects <- toxicity + No.toxicity
    no.subjects <- sum( subjects )
    txt <- "| dose levels i            |"
    output( trial.output, txt, 1:k, form, coda, sep )
    txt <- "| number of toxicities     |"
    output( trial.output, txt, toxicity, form, coda, sep )
    txt <- "| patients treated         |"
    output( trial.output, txt, subjects, form, coda, sep )
    txt <- "| % patients with toxicity |"
    chnum <- format( round( toxicity/subjects*100, 0 ) )
    output( trial.output, txt, chnum, form, coda, sep )
    txt <- "\n total percentage of observed toxicity: "
    chnum <- format( round( sum(toxicity)/no.subjects*100, 1 ) )
    coda <- ""
    sep  <- ""
    output( trial.output, txt, chnum, form, coda, sep )
    cat ("\n results are written in the file ",trial.output, "\n\n\n")
    txt <- "\n posterior mean of theta: "
    chnum <- format( round( mean.theta, 2 ) )
    output( trial.output, txt, chnum, form, coda, sep )
}

# -----------------------------------------------------------------------
#     Return the calculation models
#     ff       --> integer value:
#                    1 - hyperbolic tangent
#                    2 - logistic
#                    3 - Power
# -----------------------------------------------------------------------
which.f <- function( ff, dose, theta ) {
    return( switch(ff,
                    function(dose,theta) {(1+exp(-2*dose))^(-theta)},
                    function(dose,theta) {1/(1+exp(-3-theta*dose))},
                    function(dose,theta) {dose^exp(theta)} ) )
}

# ----------------------------------------------------------------------
#     solve the equation giving the value of x
#
#     ff         --> functional form for the probability density
#     target.tox --> desired probability of event
#     theta      --> the parameter of the model
# ----------------------------------------------------------------------
find.x <- function( ff, target.tox, theta ) {
    if ( ff == 1 ) {
        temp <- log( target.tox ) / theta
        x <- -0.5*log( exp( -temp )-1 )
    }
    else if ( ff == 2 )
        x <- -(3+log( (1-target.tox)/target.tox ))/theta
    else if ( ff == 3)
        x <- exp(log(target.tox)/exp(theta))
    return( x )
}

#-----------------------------------------------------------------------
#     Find the credible interval from the probability
#     distribution x
#
#     sdose       --> vector of doses
#     alpha       --> generate (1-alpha) of the credible level

#     x,          --> posterior distribution of mean.theta
#     mean.theta  --> mean value of x
#     ff          --> integer value choosing the functional form for the
#                 distribution of probability used in the study
#     target.tox  --> desired level of probability for toxicity
#     trial.output    --> name of the file where results are to be written
# ----------------------------------------------------------------------
cred.int <- function( sdose, alpha, x, mean.theta, ff, target.tox, trial.output ) {
    n <- length( x )
    ci.low <- 0.5*alpha
    ci.high <- 1 - ci.low
    z <- 1:200/200
    y <- quantile(x, z)    
    plot(y, z, xlab="theta", ylab="F(theta)", type="l", main="CDF" )
    psi <- which.f( ff )
    theta.low <- quantile( x, ci.low )
    segments( 0,ci.low, theta.low,ci.low )
    segments( theta.low,ci.low, theta.low,0 )    
    theta.high  <- quantile( x, ci.high )
    segments( 0,ci.high, theta.high,ci.high, lty=2 )
    segments( theta.high,ci.high, theta.high,0, lty=2 )
    txt <- c(as.character( (ci.high-ci.low)*100), "% C.I. (",
            as.character( ci.low*100 ), "% ... ",
            as.character( ci.high*100 ), "% ) =",
            format( round( theta.low, 2 ) ), "...",
            format( round( theta.high, 2 ) ))
    write("\n credible intervals for parameter theta", trial.output,
            ncol=10, append=T )
    write( txt, trial.output, ncol=10, append=T )    
    x <- seq( min( sdose ), max( sdose ), len=51 )
    main1 <- "Probability of toxicity and its credible interval"
    plot(x, psi( x, mean.theta ), ylim=c(0,1), xlab="sdose", ylab="probability of toxicity",
            type="l", main=main1, axes=F, lty=1 )
    tick <- round( as.single(sdose), 2 )
    axis(1, tick, as.character( tick ) )
    axis(2, las=1)
    box()    
    lines(x, psi( x, theta.low), lty=3 )
    lines(x, psi( x, theta.high), lty=4 )
    legend( min(x), 1, c(paste("lower 2.5%, theta=",round(theta.low,2)),
                    paste("mean ........, theta=",round(mean.theta,2)),
                    paste("upper 2.5%, theta=",round(theta.high,2))),
                    lty=c(3,1,4))
    abline( target.tox, 0, lty=2)
    min.dose <- find.x( ff, target.tox, theta.low )
    max.dose <- find.x( ff, target.tox, theta.high )
    txt[7] <- format( round( min.dose, 2 ) )
    txt[9] <- format( round( max.dose, 2 ) )
    write("\n credible intervals for MTD", trial.output,
            ncol=10, append=T )
    write( txt, trial.output, ncol=10, append=T )
}

windows(record=T)
par(mfrow=c(1,2))
p.tox0 <- c(.05,.15,.3,.45,.6)
# Hyperbolic Tangent
sdose  <- -log(1/p.tox0-1)/2.0
target.tox <-.3
crm.interactive( sdose, target.tox, ff=1 )


